/* 
  Copyright (C) 1998 Norbert Preining and Christofer Edvardsen

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "smmgr.h"

int s_fax = 1;
int s_voice = 1;
int s_monitor = 1;
int s_remote = 0;
int s_answer = 4;
int s_password = 1234;
char s_identity[21] = "";
char s_device[32] = "ttyS1";
speed_t s_baud = B38400;
int s_verbose = 1;
char s_init[32] = "AT&F1";
char s_faxdir[256];
char s_voicedir[256];
char s_libfile[256];
char s_logfile[256];
char s_lockdir[256];
char s_configfile[256];
char s_faxprefix[256];
int s_pagesingle = 1;
int s_ignore_dle = 0;
int s_imon_on_exit = 0;
int s_mail_on_download = 0;
int s_out_card = 0;
int c_firstfree = -1;
int s_recordtime = 0;
char s_faxmail[256], s_voicemail[256], s_mailprog[256], s_viewprog[256];
char s_localscript[256];
int index2fnr[256];
char current_messagename[256];
int s_modemtype=0;
int s_tollsaver=0, s_loudness=0;
int getblock_return;

msginfo *msginfos;

int nr_msgs = 0;
int fd=-1, nr_voice, nr_fax, nr_voice_new, nr_fax_new, device_initialized=0;
int percentfull = 0,inmsg=0,fullmsg=0, total_memory=0;
int c_page = -1, c_offset, have_msginfos = 0;
int page_size;
Tpage pages[64];
unsigned char *page;
time_t stored_time = -1, timedrift;
char lockname[32]="";
char tmpdir[256];
char rmcommand[300];

static char *line_read = (char *)NULL;


struct termios oldiosettings, iosettings;

Tmodemdef modemtypes[] = {
	{ "MessagePlus",	USR_MESSAGE_PLUS },
	{ "MessagePro",		USR_MESSAGE_PRO	 },
	{ NULL, 0 }
};

Tmodemstat modemstat;

#ifdef MACRO
Tmacdef *macdef;
#endif

Setting comdef_settings[] = {
	{ "modemtype",	
	  "Set Modem Type",		
	   ALL_MODEMS,
	  S_INT,  &s_modemtype, &set_modemtype },
	{ "faxdir",	
	  "set directory where faxes are saved",
	  ALL_MODEMS,
	  S_CHAR, &s_faxdir[0], &set_faxdir },
	{ "voicedir",	
	  "set directory where voice are saved",
	  ALL_MODEMS,
	  S_CHAR, &s_voicedir[0], &set_voicedir },
	{ "fax",	
	  "activate fax reception in SelfMode",	
	  ALL_MODEMS,
	  S_BOOL, &s_fax, &set_fax },
	{ "voice",	
	  "activate voice reception in SelfMode",
	  ALL_MODEMS, 
	  S_BOOL, &s_voice, &set_voice },
	{ "monitor",	
	  "activate output to speaker while recording",
	  ALL_MODEMS, 
	  S_BOOL, &s_monitor, &set_monitor },
	{ "remote",	
	  "activate remote control",	
	  ALL_MODEMS,
	  S_BOOL, &s_remote, &set_remote },
	{ "answer",	
	  "set # of rings",		ALL_MODEMS,	
	  S_INT,  &s_answer, &set_answer },
	{ "password",	
	  "set remote passwd",		ALL_MODEMS, 
	  S_INT,  &s_password, &set_password },
	{ "identity",	
	  "set fax identity",		ALL_MODEMS, 
	  S_CHAR, &s_identity[0], &set_identity},
	{ "device",	
	  "set device to use",			ALL_MODEMS, 
	  S_CHAR, &s_device[0], &set_device},
	{ "verbose",	
	  "adjusts verbosity level (silent,error,info,debug)",
	  ALL_MODEMS,
	  S_INT,  &s_verbose, &set_verbose},
	{ "initstring",	
	  "set string send to modem at initialization",	
	  ALL_MODEMS,
	  S_CHAR, &s_init[0], &set_initstring},
	{ "libfile",	
	  "specifies the file to save timing-data to",
	  ALL_MODEMS, 
	  S_CHAR, &s_libfile[0], &set_libfile},
	{ "logfile",	
	  "specifies the file to write log-entries to",
	  ALL_MODEMS,
	  S_CHAR, &s_logfile[0], &set_logfile},
	{ "lockdir",	
	  "specifies where to put lock-files",
	  ALL_MODEMS, 
	  S_CHAR, &s_lockdir[0], &set_lockdir},
	{ "ignoredle",   
	  "set ignore_dle flag", 	ALL_MODEMS, 
	  S_BOOL, &s_ignore_dle, &set_ignore_dle}, 
	{ "out_card",   
	  "play to soundcard or to internal speaker",
	  ALL_MODEMS,
	  S_BOOL, &s_out_card, &set_out_card},
	{ "imon_on_exit", 
	  "Run imon on exit",   	ALL_MODEMS, 
	  S_BOOL, &s_imon_on_exit, &set_imon_on_exit}, 
	{ "configfile",  
	  "specify configfile for next `parsefile'",
	  ALL_MODEMS,	
	  S_CHAR, &s_configfile[0], &set_configfile},
	{ "mailprog",   
	  "mail program cmd line",	ALL_MODEMS, 
	  S_CHAR, &s_mailprog[0], &set_mailprog},
	{ "viewprog",    
	  "view fax program cmd line", 	ALL_MODEMS, 
	  S_CHAR, &s_viewprog[0], &set_viewprog},
	{ "faxmail",   
	  "adress to send faxes to",	ALL_MODEMS, 
	  S_CHAR, &s_faxmail[0], &set_faxmail},
	{ "voicemail",   
	  "adress to send voices to",  	ALL_MODEMS, 
	  S_CHAR, &s_voicemail[0], &set_voicemail},
        { "mail_on_download", 
	  "mail msg after download",   	ALL_MODEMS, 
	  S_BOOL, &s_mail_on_download, &set_mail_on_download},
	{ "recordtime", 
	  "Set record time limit in sec (0 or 255 disables)", ALL_MODEMS,
	  S_INT, &s_recordtime, &set_recordtime}, 
	{ "localscript", 
	  "cmd for editlocal",   ALL_MODEMS,
	  S_CHAR, &s_localscript[0], &set_localscript},
	{ "loudness",
	  "set loudness of rings",	USR_MESSAGE_PRO,
	  S_INT, &s_loudness,  &set_loudness},
	{ "tollsaver",
	  "activate tollsaving",     USR_MESSAGE_PRO,
	  S_BOOL, &s_tollsaver, &set_tollsaver},
	{ NULL, NULL, S_CHAR, 0, NULL }
};
	

Tcomdef comdef[] = {
	{ "imon",	
	  "Start SelfMode",	
	  &repl_imon,			0,		ALL_MODEMS},
	{ "imon_proper",
	  "Start SelfMode",
	  &repl_imon_proper, 		0,		ALL_MODEMS},
	{ "imoff",	
	  "Stop SelfMode",	
	  &repl_imoff, 			0, 		ALL_MODEMS},
	{ "getmsginfo",	
	  "get msgs infos",	
	  &repl_getmsginfo, 		0, 		ALL_MODEMS},
	{ "get",        
	  "get msgs <number>|all|new|newvoice|newfax|greeting|full", 
	  &repl_get,  			1, 		ALL_MODEMS},
	{ "clear",	
	  "clear modem memory",		
	  &repl_clear, 			0, 		ALL_MODEMS},
	{ "initdevice",	
	  "init device",		
	  &repl_initdevice, 		0, 		ALL_MODEMS},
	{ "closedevice", 
	  "close device",	
	  &repl_closedevice, 		0, 		ALL_MODEMS},
	{ "ulg",	
	  "upload greeting msg from <file>",	
	  &repl_ulg,	   		1, 		ALL_MODEMS},
	{ "ulf",	
	  "upload full msg from <file>",		
	  &repl_ulf,	   		1, 		ALL_MODEMS},
	{ "recg",	
	  "record greeting msg",	
	  &repl_recg,	   		0, 		ALL_MODEMS},
	{ "recf",	
	  "record full msg",		
	  &repl_recf,			0, 		ALL_MODEMS},
	{ "play",	
	  "play voice msg <number>|all|new|greeting|full",	
	  &repl_play, 			1, 		ALL_MODEMS},
	{ "view",	
	  "view fax <number>|all|new",		
	  &repl_view,			1, 		ALL_MODEMS},
	{ "list",	
	  "list msgs",		
	  &repl_list,			0, 		ALL_MODEMS},
	{ "report",	
	  "report",		
	  &repl_report,			0, 		ALL_MODEMS},
	{ "clock",	
	  "read clock",		
	  &repl_clock,			0, 		ALL_MODEMS},
	{ "parsefile",	
	  "parse config file",	
	  &repl_parseconfig, 		0, 		ALL_MODEMS},
	{ "freenr",	
	  "freenr",		
	  &repl_freenr,			0, 		ALL_MODEMS},
	{ "repl",	
	  "start repl",		
	  &repl_repl,			0, 		ALL_MODEMS},
	{ "quit",	
	  "quit repl",		
	  &repl_quit,			0, 		ALL_MODEMS},
	{ "exit",	
	  "quit repl",		
	  &repl_quit,			0, 		ALL_MODEMS},
	{ "bye",	
	  "quit repl",		
	  &repl_quit,			0, 		ALL_MODEMS},
/* 	{ "--help",	  */
/* 	  "get help",		 */
/* 	  &repl_help,			0, 		ALL_MODEMS}, */
/* 	{ "-help",	  */
/* 	  "get help",		 */
/* 	  &repl_help,			0, 		ALL_MODEMS}, */
/* 	{ "-h",		  */
/* 	  "get help",		 */
/* 	  &repl_help, 			0, 		ALL_MODEMS}, */
	{ "help",	 
	  "get help",		
	  &repl_help,	   		0, 		ALL_MODEMS},
	{ "help_settings",	 
	  "get help about settings",		
	  &repl_helpsettings,  		0, 		ALL_MODEMS},
	{ "version",     
	  "display version",     
	  &repl_version,     		0, 		ALL_MODEMS},
	{ "settings",    
	  "List settings",       
	  &repl_settings,    		0, 		ALL_MODEMS},
	{ "modemstatus", 
	  "List modem status",   
	  &repl_modemstatus, 		0, 		ALL_MODEMS},
	{ "savepage",    
	  "Save page <n> to file", 
	  &repl_savepage,  		1, 		ALL_MODEMS}, 
	{ "editlocal",	
	  "edit local messages",	
	  &repl_editlocal,		0,		ALL_MODEMS},
	{ "command",    
	  "send <command> to modem", 
	  &repl_command,    		1, 		ALL_MODEMS},
	{ "set",        
	  "change <<setting>>",       
	  &repl_set,         		2, 		ALL_MODEMS},
	{ "showsetting",
	  "show <setting>",         
	  &repl_showsetting, 		1, 		ALL_MODEMS},
#ifdef MACRO
	{ "macro",      
	  "defines a <<macro>>",      
	  &repl_macro,       		2, 		ALL_MODEMS},
	{ "showmacros", 
	  "show defined macros", 
	  &repl_showmacros,   		0, 		ALL_MODEMS},
#endif
	{ NULL, NULL, NULL, 0 , 0}
};


/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *
rl_gets ()
{
  /* If the buffer has already been allocated, return the memory
     to the free pool. */
  if (line_read)
    {
      free (line_read);
      line_read = (char *)NULL;
    }

  /* Get a line from the user. */
  line_read = readline (PROMPT);

  /* If the line has any text in it, save it on the history. */
  if (line_read && *line_read)
    add_history (line_read);

  return (line_read);
}

int main(int argc, char **argv) {
	int g_argc, c_arg;
	char **g_argv;
	int i;
	
	strncpy(s_configfile,CONFIGFILE,256);
	strncpy(s_faxdir,SPOOLDIR,256);
	strncpy(s_voicedir,SPOOLDIR,256);
	strncpy(s_faxprefix,FAXPREFIX,256);
	snprintf(s_libfile,256,"%s/%s",LIBDIR,"reset_time");
	strncpy(s_logfile,"",256);
	strncpy(s_lockdir,LOCKDIR,256);
	for (i=0;i<64;i++) {
		index2fnr[i]=-1;
		pages[i].data = NULL;
		pages[i].filled = 0;
		pages[i].length = 0;
	}
	snprintf(tmpdir,256,"/tmp/smmgr-%d",getpid());
	snprintf(rmcommand,300,"rm -rf %s",tmpdir);
	
	on_exit(exitfunc,(void*)NULL);
	if (signal (SIGINT, termination_handler) == SIG_IGN) signal (SIGINT, SIG_IGN);
	if (signal (SIGHUP, termination_handler) == SIG_IGN) signal (SIGHUP, SIG_IGN);
	if (signal (SIGTERM, termination_handler) == SIG_IGN) signal (SIGTERM, SIG_IGN);

#ifdef MACRO
	macdef = (Tmacdef*)malloc(sizeof(Tmacdef));
	macdef->name = NULL;
	macdef->expansion = NULL;
	macdef->next = NULL;
#endif

	g_argv = argv;
	g_argc = argc;
	c_arg = 1;

	if (argc == 2 && (strcmp(argv[1],"help") == 0 ||
			  strcmp(argv[1],"--help") == 0 ||
			  strcmp(argv[1],"-help") == 0 ||
			  strcmp(argv[1],"-h") == 0)) {
		repl_helpall(NULL);
		exit(0);
	}
	if (argc == 2 && (strcmp(argv[1],"--version") == 0 ||
			  strcmp(argv[1],"-version") == 0 ||
			  strcmp(argv[1],"-v") == 0 ||
			  strcmp(argv[1],"version") == 0)) {
		repl_version(NULL);
		exit(0);
	}
	if (argc > 1 && strcmp(argv[1],"noconfig") == 0) {
		logit(L_INFO,"Skip reading configfile!\n");
		c_arg++;
	} else {
		if (parseconfigfile() == E_FILE) {
			logit(L_ERROR,"Error parsing configfile!\n");
			logit(L_ERROR,"Maybe you should specify \"noconfig\" as first argument!?\n");
			exit(1);
		}
	}

	initialize_readline();
	if (c_arg < g_argc) {
		return(start_processing(g_argc, g_argv, &c_arg));
	} else {
		return(start_repl());
	}
	return(0);
}

int start_repl() {
	int ret = 0;
	int g_argc, c_arg;
	char **g_argv;
	char *rl;

	while ((rl = rl_gets()) != NULL) {
		if (is_not_comment(rl)) {
			g_argc = line_to_array(rl,&g_argv);
			if (g_argc > 0) {
				c_arg = 0;
				ret += start_processing(g_argc,g_argv,&c_arg);
				free(g_argv);
			}
		}
	}
	return(ret);
}

int start_processing(int g_argc, char **g_argv, int *c_arg) {
	int ret=0;
	while ( *c_arg < g_argc ) ret += process_one(g_argc,g_argv,c_arg);
	return(ret);
}

int process_one(int g_argc, char **g_argv, int *c_arg) {
	Tcomdef *cd = &comdef[0];
#ifdef MACRO
	char **l_argv;
	Tmacdef *md = macdef;
	int lc_arg, ret, l_argc;
	char tmp[256];
#endif
	while (cd->name != NULL ) {
		if (strcasecmp( g_argv[*c_arg], cd->name ) == 0) {
			(*c_arg)++;
			/* check on modemtype */
			if (!(cd->whichmodems & modemtypes[s_modemtype].id)) {
				if (cd->takearg) {
					if (*c_arg + cd->takearg > g_argc) {
						logit(L_ERROR,"Command requires %d argument(s): %s\n",cd->takearg,g_argv[(*c_arg)-1]);
					} else {
						(*c_arg) += cd->takearg;
					}
				}
				logit(L_ERROR,"This command is not allowed for modemtype %s!\n",modemtypes[s_modemtype].modemname);
				return(P_IGNORED);
			}
			if (cd->takearg) {
				if (*c_arg + cd->takearg > g_argc) {
					logit(L_ERROR,"Command requires %d argument(s): %s\n",cd->takearg,g_argv[(*c_arg)-1]);
					return(P_NOTFOUND);
				}
				ret = (*(cd->pfun))(&(g_argv[*c_arg]));
				*(c_arg) += cd->takearg;
				return(ret);
			} else {
				return((*(cd->pfun))(NULL));
			}
		}
		cd++;
	}
#ifdef MACRO
	/* now check on macro definition */
	logit(L_INFO,"Searching for macro: %s\n",g_argv[*c_arg]);
	while (md->name != NULL ) {
		logit(L_DEBUG,"md->name=%s\n",md->name);
		if (strcasecmp( g_argv[*c_arg], md->name ) == 0) {
			ret = 0;
			strcpy(tmp,md->expansion);
			l_argc = line_to_array(tmp,&l_argv);
			if (l_argc > 0) {
				lc_arg = 0;
				ret = start_processing(l_argc,l_argv,&lc_arg);
				free(l_argv);
				(*c_arg)++;
				return(ret);
			}
			
		}
		md = md->next;
	}
#endif /* MACRO	*/
	logit(L_ERROR,"Command not found: %s\n",g_argv[*c_arg]);
	(*c_arg)++;
	return(P_NOTFOUND);
}

int parseconfigfile() {
	char s[256];
	int ret = 0;
	int g_argc, c_arg;
	char **g_argv;
	FILE *cfile;
	
	cfile = fopen(s_configfile,"r");
	if (cfile == NULL) {
		logit(L_ERROR,"Couldn't open config file for reading!\n");
		return(E_FILE);
	}
	while (fgets(s,256,cfile) != NULL) {
		if (is_not_comment(s)) {
			g_argc = line_to_array(s,&g_argv);
			if (g_argc > 0) {
				c_arg = 0;
				ret += start_processing(g_argc,g_argv,&c_arg);
				free(g_argv);
			}
		}
	}
	return(ret);
}


/* 
   exitfunc
*/
void exitfunc(int n,void *p) {
	int i;
	if (n==0 && s_imon_on_exit) repl_imon(NULL);
	logit(L_INFO,"Restoring old io settings and removing lockfile.\n");
	if (have_msginfos) {
		free(msginfos);
		have_msginfos = 0;
	}
	for (i=0;i<64;i++) {
		if (pages[i].filled) free(pages[i].data);
	}
	repl_closedevice(NULL);
	/* clean up tmp dir */
	system(rmcommand);
} /* exitfunc */

void termination_handler (int sig) {
	logit(L_ERROR,"Received signal %d -- ooops\n",sig);
	exit(1); 
}

void interrupt_play_handler (int sig) {
	logit(L_DEBUG,"Got signal %d.\nSending interrupt to modem!\n");
	write(fd,"%c",3);
}	
void writetolog(char *writestr) {
	FILE *logfile;
	time_t *timenow;
	char *datestr;

	if ( strlen(s_logfile)>0  ) {
		logfile = fopen(s_logfile,"a");
		if ( logfile == NULL ) {
			logit(L_ERROR,"Couldn't write to logfile: %s\n",s_logfile);
		} else {
			timenow=(time_t*)malloc(sizeof(timenow));
			*timenow=time(NULL);
			datestr=ctime( timenow );
			datestr[strlen(datestr)-1]=0;
			fprintf(logfile,"%s %s\n",datestr,writestr);
			free(timenow);
		}
	}
} /* writetolog */


/* **************************************************************** */
/*                                                                  */
/*                  Interface to Readline Completion                */
/*                                                                  */
/* **************************************************************** */
     
/* Tell the GNU Readline library how to complete.  We want to try to complete
   on command names if this is the first word in the line, or on filenames
   if not. */
void initialize_readline () {
	/* Allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "smmgr";
	rl_completion_entry_function = (rl_compentry_func_t *)command_generator;
}
     
/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
char *command_generator (const char *text, int state) {
	static int list_index, len;
	char *name;
     
	/* If this is a new word to complete, initialize now.  This includes
	   saving the length of TEXT for efficiency, and initializing the index
	   variable to 0. */
	if (!state) {
		list_index = 0;
		len = strlen (text);
	}
     
	/* Return the next name which partially matches from the command list. */
	while ((name = comdef[list_index].name)) {
		list_index++;
		
		if (strncmp (name, text, len) == 0)
			return (dupstr(name));
	}
     
	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}


